exclude(:test_expand_heap, "testing behaviour specific to default GC")
exclude(:test_gc_config_disable_major, "testing behaviour specific to default GC")
exclude(:test_gc_config_disable_major_gc_start_always_works, "testing behaviour specific to default GC")
exclude(:test_gc_config_full_mark_by_default, "testing behaviour specific to default GC")
exclude(:test_gc_config_invalid_args, "testing behaviour specific to default GC")
exclude(:test_gc_config_setting_returns_updated_config_hash, "testing behaviour specific to default GC")
exclude(:test_gc_internals, "testing behaviour specific to default GC")
exclude(:test_gc_parameter, "testing behaviour specific to default GC")
exclude(:test_gc_parameter_init_slots, "testing behaviour specific to default GC")
exclude(:test_heaps_grow_independently, "testing behaviour specific to default GC")
exclude(:test_latest_gc_info, "testing behaviour specific to default GC")
exclude(:test_latest_gc_info_argument, "testing behaviour specific to default GC")
exclude(:test_latest_gc_info_need_major_by, "testing behaviour specific to default GC")
exclude(:test_latest_gc_info_weak_references_count, "testing behaviour specific to default GC")
exclude(:test_old_to_young_reference, "testing behaviour specific to default GC")
exclude(:test_profiler_enabled, "MMTk does not have GC::Profiler")
exclude(:test_profiler_raw_data, "MMTk does not have GC::Profiler")
exclude(:test_profiler_total_time, "MMTk does not have GC::Profiler")
exclude(:test_start_full_mark, "testing behaviour specific to default GC")
exclude(:test_start_immediate_sweep, "testing behaviour specific to default GC")
exclude(:test_stat, "testing behaviour specific to default GC")
exclude(:test_stat_constraints, "testing behaviour specific to default GC")
exclude(:test_stat_heap, "testing behaviour specific to default GC")
exclude(:test_stat_heap_all, "testing behaviour specific to default GC")
exclude(:test_stat_heap_constraints, "testing behaviour specific to default GC")
exclude(:test_thrashing_for_young_objects, "testing behaviour specific to default GC")
