require_relative '../../../spec_helper'
require 'net/http'

describe "Net::HTTP#close_on_empty_response" do
  it "needs to be reviewed for spec completeness"
end

describe "Net::HTTP#close_on_empty_response=" do
  it "needs to be reviewed for spec completeness"
end
