#if USE_ZJIT

/* insn <%= insn.pretty_name %> */
INSN_ENTRY(<%= insn.name %>)
{
    START_OF_ORIGINAL_INSN(<%= insn.name %>);
    rb_zjit_profile_insn(BIN(<%= insn.jump_destination %>), ec);
    DISPATCH_ORIGINAL_INSN(<%= insn.jump_destination %>);
    END_INSN(<%= insn.name %>);
}

#endif
