/* -*- C -*- */

%# Copyright (c) 2017 Urabe, Shyouhei.  All rights reserved.
%#
%# This file is a part of  the programming language Ruby.  Permission is hereby
%# granted, to either  redistribute and/or modify this file,  provided that the
%# conditions mentioned  in the  file COPYING  are met.   Consult the  file for
%# details.
<%= render 'copyright' %>
<%= render 'notice', locals: {
  this_file: 'contains instruction information for yarv instruction sequence.',
  edit: __FILE__,
} %>
#ifndef INSNS_INFO_INC
#define INSNS_INFO_INC 1
<%= render 'insn_type_chars' %>
<%= render 'insn_name_info' %>
<%= render 'insn_len_info' %>
<%= render 'insn_operand_info' %>
<%= render 'leaf_helpers' %>
<%= render 'sp_inc_helpers' %>
<%= render 'zjit_helpers' %>
<%= render 'attributes' %>
<%= render 'insn_leaf_info' %>
<%= render 'comptime_insn_stack_increase' %>
#endif
